@@ -36,11 +36,29 @@ class TestSSLConnection(unittest.TestCase):
     @asyncio.coroutine
     def setUp(self):
         self.loop = asyncio.get_event_loop()
+        self.server = None
         self.server_ctx = ssl.create_default_context(
             ssl.Purpose.CLIENT_AUTH
         )
 
         self.server_ctx.load_cert_chain(str(KEYFILE))
+        yield from self._replace_server()
+
+        self.inbound_queue = asyncio.Queue()
+
+    @asyncio.coroutine
+    def _shutdown_server(self):
+        self.server.close()
+        while not self.inbound_queue.empty():
+            reader, writer = yield from self.inbound_queue.get()
+            writer.close()
+        yield from self.server.wait_closed()
+        self.server = None
+
+    @asyncio.coroutine
+    def _replace_server(self):
+        if self.server is not None:
+            yield from self._shutdown_server()
 
         self.server = yield from asyncio.start_server(
             self._server_accept,
@@ -48,16 +66,11 @@ class TestSSLConnection(unittest.TestCase):
             port=PORT,
             ssl=self.server_ctx,
         )
-        self.inbound_queue = asyncio.Queue()
 
     @blocking
     @asyncio.coroutine
     def tearDown(self):
-        self.server.close()
-        while not self.inbound_queue.empty():
-            reader, writer = yield from self.inbound_queue.get()
-            writer.close()
-        yield from self.server.wait_closed()
+        yield from self._shutdown_server()
 
     def _server_accept(self, reader, writer):
         self.inbound_queue.put_nowait(
@@ -67,6 +80,7 @@ class TestSSLConnection(unittest.TestCase):
     def _stream_reader_proto(self):
         reader = asyncio.StreamReader(loop=self.loop)
         proto = asyncio.StreamReaderProtocol(reader)
+        proto.aioopenssl_test_reader = reader
         return proto
 
     def _connect(self, *args, **kwargs):
@@ -77,7 +91,8 @@ class TestSSLConnection(unittest.TestCase):
                 *args,
                 **kwargs
             )
-        reader = reader_proto._stream_reader
+        reader = reader_proto.aioopenssl_test_reader
+        del reader_proto.aioopenssl_test_reader
         writer = asyncio.StreamWriter(transport, reader_proto, reader,
                                       self.loop)
         return transport, reader, writer
@@ -321,17 +336,22 @@ class TestSSLConnection(unittest.TestCase):
     @blocking
     @asyncio.coroutine
     def test_renegotiation(self):
-        def ctx_factory(_):
-            ctx = OpenSSL.SSL.Context(
-                OpenSSL.SSL.SSLv23_METHOD
-            )
-            ctx.set_options(getattr(ssl, "OP_NO_TLSv1_3", 0))
-            return ctx
+        self.server_ctx = ssl.create_default_context(
+            ssl.Purpose.CLIENT_AUTH
+        )
+        if hasattr(ssl, "OP_NO_TLSv1_3"):
+            # Need to forbid TLS v1.3, since TLSv1.3+ does not support
+            # renegotiation
+            self.server_ctx.options |= ssl.OP_NO_TLSv1_3
+        self.server_ctx.load_cert_chain(str(KEYFILE))
+        yield from self._replace_server()
 
         c_transport, c_reader, c_writer = yield from self._connect(
             host="127.0.0.1",
             port=PORT,
-            ssl_context_factory=ctx_factory,
+            ssl_context_factory=lambda transport: OpenSSL.SSL.Context(
+                OpenSSL.SSL.SSLv23_METHOD
+            ),
             server_hostname="localhost",
             use_starttls=False,
         )
@@ -424,12 +444,16 @@ class TestSSLConnection(unittest.TestCase):
             )
 
         exc = ctx.exception
-
-        self.assertTrue(type(exc) is asyncio.streams.IncompleteReadError or
-                        type(exc) is ConnectionResetError)
-
-        if isinstance(exc, asyncio.streams.IncompleteReadError):
+        # using type(None) as default, since that will always be false in the
+        # isinstance check below
+        incomplete_read_exc_type = getattr(
+            asyncio.streams, "IncompleteReadError",
+            getattr(asyncio, "IncompleteReadError", type(None))
+        )
+        if isinstance(exc, incomplete_read_exc_type):
             self.assertFalse(exc.partial)
+        elif not isinstance(exc, ConnectionResetError):
+            raise exc
 
     @blocking
     @asyncio.coroutine
@@ -634,11 +658,20 @@ class TestSSLConnectionThreadServer(unittest.TestCase):
     @asyncio.coroutine
     def setUp(self):
         self.loop = asyncio.get_event_loop()
+        self.thread = None
 
         ctx = OpenSSL.SSL.Context(OpenSSL.SSL.SSLv23_METHOD)
         ctx.use_certificate_chain_file(str(KEYFILE))
         ctx.use_privatekey_file(str(KEYFILE))
 
+        self._replace_thread(ctx)
+
+    def _replace_thread(self, ctx):
+        if self.thread is not None:
+            self.thread.stopped = True
+            self.thread.join()
+            self.thread = None
+
         self.inbound_queue = asyncio.Queue()
         self.thread = ServerThread(
             ctx,
@@ -680,6 +713,7 @@ class TestSSLConnectionThreadServer(unittest.TestCase):
     def _stream_reader_proto(self):
         reader = asyncio.StreamReader(loop=self.loop)
         proto = asyncio.StreamReaderProtocol(reader)
+        proto.aioopenssl_test_reader = reader
         return proto
 
     def _connect(self, *args, **kwargs):
@@ -690,7 +724,8 @@ class TestSSLConnectionThreadServer(unittest.TestCase):
                 *args,
                 **kwargs
             )
-        reader = reader_proto._stream_reader
+        reader = reader_proto.aioopenssl_test_reader
+        del reader_proto.aioopenssl_test_reader
         writer = asyncio.StreamWriter(transport, reader_proto, reader,
                                       self.loop)
         return transport, reader, writer
@@ -737,17 +772,23 @@ class TestSSLConnectionThreadServer(unittest.TestCase):
     @blocking
     @asyncio.coroutine
     def test_renegotiate(self):
-        def ctx_factory(_):
-            ctx = OpenSSL.SSL.Context(
-                OpenSSL.SSL.SSLv23_METHOD
-            )
-            ctx.set_options(getattr(ssl, "OP_NO_TLSv1_3", 0))
-            return ctx
+        ctx = OpenSSL.SSL.Context(OpenSSL.SSL.SSLv23_METHOD)
+        # Need to forbid TLS v1.3, since TLSv1.3+ does not support
+        # renegotiation
+        ctx.set_options(
+            getattr(OpenSSL.SSL, "OP_NO_TLSv1_3",
+                    getattr(ssl, "OP_NO_TLSv1_3", 0))
+        )
+        ctx.use_certificate_chain_file(str(KEYFILE))
+        ctx.use_privatekey_file(str(KEYFILE))
+        self._replace_thread(ctx)
 
         c_transport, c_reader, c_writer = yield from self._connect(
             host="127.0.0.1",
             port=PORT+1,
-            ssl_context_factory=ctx_factory,
+            ssl_context_factory=lambda transport: OpenSSL.SSL.Context(
+                OpenSSL.SSL.SSLv23_METHOD
+            ),
             server_hostname="localhost",
             use_starttls=False,
         )
@@ -774,7 +815,17 @@ class TestSSLConnectionThreadServer(unittest.TestCase):
             b"fnord"
         )
 
-        sock.renegotiate()
+        try:
+            sock.renegotiate()
+        except OpenSSL.SSL.Error as exc:
+            (argv,), = exc.args
+            if (argv[1] == "SSL_renegotiate" and
+                    argv[2] == "wrong ssl version"):
+                raise RuntimeError(
+                    "You are a PyOpenSSL version which uses TLSv1.3, but has"
+                    " no way to turn it off. Update PyOpenSSL."
+                )
+            raise
 
         c_writer.write(b"baz")
 
